#!/usr/bin/env python

from sa import utils, repo as rp
import os, re, pathlib, json
import tempfile
import shutil
def check_prev_commit(repo, custom_tags):
    """Check to make sure the previous commit is not an automated on or contains specific tags"""
    return None

def version_get(files, options):

    build_info = []
    # Regex pattern to match all version numbers
    version_num_pattern = re.compile('(\\d*\\d?\\.\\d*\\d)+')
    for f in files:
        build_info = build_info + parse_version_file(f, version_num_pattern, options=options)
    return build_info

def version_bump(repo, files, options):
    # If repo is valid, get the most recent commit hash
    # short_hash = repo.git.rev_parse('HEAD', short=True)
    if options.git_username != None:
        git_username = options.git_username
    else:
        git_username = 'SourceAssist'
    if options.git_useremail != None:
        git_useremail = options.git_useremail
    else:
        git_useremail = 'SourceAssist@noreply.com'


    # Regex pattern to match all version numbers
    version_num_pattern = re.compile('(\\d*\\d?\\.\\d*\\d)+')
    build_info_string = '[git-version-bump]'
    for f in files:
        build_info = parse_version_file(f, version_num_pattern, write=True, repo=repo, options=options)
        for file, version in build_info:
            build_info_string = ''.join([build_info_string,'\n', file, ' ', version])
        repo.git.add(f)


    repo.git.config(['user.name', [git_username]])    
    repo.git.config(['user.email', [git_useremail]])    
    repo.git.commit('-m', build_info_string)
    #TODO: fix multi-build number issues
    # repo.create_tag(''.join(['v',build_num]), message=f'Automatic tag \"v{build_num}\"')

def update_version(prev_ver_num):
    version_num = prev_ver_num.split('.')
    build_num = int(version_num[-1])
    build_num += 1

    new_version_num = ''
    for num in version_num[:-1]:
        new_version_num = ''.join([new_version_num, str(num), '.'])
    
    new_version_num = ''.join([new_version_num, str(build_num)])
    return new_version_num

def version_data_from_dict(dict_obj, keys, file_names, write=False):

    version_data = []
    found_dicts = utils.find_key_pairs(dict_obj, keys)

    for found_dict in found_dicts:

        # only update version number for files that have been modified
        if any([found_dict['file'].lower() in file_name.lower() for file_name in file_names]):
            print(f'Found on changes in: {found_dict["file"]} {found_dict["version"]}')
            # update version number if specified
            found_dict['version'] = update_version(found_dict['version'])
            print(f'Updated to: {found_dict["file"]} {found_dict["version"]}')
            # write and return only updated version numbers
            version_data.append((found_dict['file'], found_dict['version']))
        
        elif not write:
            # return all version data found
            version_data.append((found_dict['file'], found_dict['version']))
        
    return version_data

def parse_version_file(file_path, regex_pattern, write=False, **kwargs):
    """Bumps up the build version number of the specified file"""

    options = None
    repo = None
    if 'options' in kwargs:
        options = kwargs.get('options')
    if 'repo' in kwargs:
        repo = kwargs.get('repo')

    version_info = []
    new_version_num = ''
    new_file = None

    with open(file_path, 'r') as old_file:
        if write:
            #Create temp file
            fh, abs_path = tempfile.mkstemp()
            new_file = os.fdopen(fh, 'w')

        # check if JSON, and if the file is a docker versioning JSON
        if options != None and 'docker' in options and options.docker:
            changed_files = rp.get_changed_files(repo)
            docker_info = json.load(old_file)
            version_info = version_info + version_data_from_dict(docker_info, ['version','file'], [str(path) for path in changed_files], write=write)
            if write:
                json.dump(docker_info, new_file, indent=4)

        else:
            for i, line in enumerate(old_file):
                
                matches = re.finditer(regex_pattern, line)
                if not is_valid_version(file_path, line):
                    if write:
                        new_file.write(line)
                    continue

                for match in matches:
                    prev_ver_num = match.group()
                    new_version_num = update_version(prev_ver_num)
                    version_info.append((os.path.basename(file_path), new_version_num))
                    new_line = line.replace(prev_ver_num, new_version_num)
                    if write:
                        new_file.write(new_line)
                        print(f'Found on line {i+1}: {prev_ver_num}')
                        print(f'Changed {line.strip()} to {new_line.strip()}')
    if write:
        new_file.close()
        #Copy the file permissions from the old file to the new file
        shutil.copymode(file_path, abs_path)
        #Remove original file
        os.remove(file_path)
        #Move new file
        shutil.move(abs_path, file_path)

    return version_info



def is_valid_version(file_path, line):
    """Returns if the specified line that contains a version number
       is the valid one for that file type"""

    file_name = os.path.basename(file_path)
    ext = pathlib.Path(file_path).suffix
    if ext != '':
        if ext == '.py':
            if '__version__' in line:
                return True
        
        if ext == '.json':
            if 'version' in line:
                return True
        
        if ext == '.csproj':
            if '</Version>' in line:
                return True
        
        if ext == '.cs':
            if 'AssemblyVersion' in line or 'AssemblyFileVersion' in line:
                return True
    else:
        if file_name == 'Doxyfile':
            if 'PROJECT_NUMBER' in line:
                return True
    
    return False



#!/usr/bin/env python

import json

def get_data_from_json(file):
    with open(file, 'r') as json_file:
        return json.load(json_file)


def find_key_pairs(obj, keys, found_pairs=[]):
    """Traverses a nested dictionary and lists to find a key, or list of key pairs and returns all parent dictionaries
        that contain the key/key pairs"""
    if not isinstance(keys, list):
        keys = [keys]

    if isinstance(obj, list):
        for obj in obj:
            find_key_pairs(obj, keys, found_pairs)
    elif isinstance(obj, dict):
        # check if all keys are in dictionary
        if not(keys - obj.keys()):
            # only update version number for files that have been modified
            found_pairs.append(obj)
        # check for a nested dictionary or list of dictionaries
        else:
            for k,v in obj.items():
                if isinstance(v, (dict, list)):
                    find_key_pairs(v, keys, found_pairs)

    return found_pairs



#!/usr/bin/env python

import os, git

# push code from an https cloned repo
def push(repo, username, password):

    for remote in repo.remotes:
        index = remote.url.find('//') + 2
        new_url = remote.url[:index] + f'{username}:{password}@' + remote.url[index:]
        remote.set_url(new_url)

    os.environ['GIT_USERNAME'] = username
    os.environ['GIT_PASSWORD'] = password
    for remote in repo.remotes:
        remote.push()
        remote.push(['--tag'])

def reattach_head(repo):
    head_info = repo.git.show(['-s', '--pretty=%d', 'HEAD'])
    for branch in repo.refs:
        branch = str(branch).replace('origin/','')
        if branch in head_info:
            repo.git.checkout(branch)
            break

def is_git_repo(path):
    """Returns if a given path is a git repo"""

    try:
        _ = git.Repo(path).git_dir
        return True
    except Exception:
        return False

def get_changed_files(repo):
    """return a list of all of the changed files based on the last commit"""
    if repo != None:
        return list(repo.commit(repo.head.object.hexsha).stats.files.keys())
    return []



from sa._version import __version__
from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="Source Assist",
    version=__version__,
    author="Kyle Petryszak",
    author_email="",
    description="A DevOps automation tool for incrementing version numbers and git repo management",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/projectinitiative/SourceAssist",
    project_urls={"Bug Tracker": "https://github.com/projectinitiative/SourceAssist/issues",},
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    packages=find_packages(),
    install_requires=[
        "gitpython",
        "docker"
    ],
    python_requires=">=3.6",
    entry_points={"console_scripts": ["sa = sa.source_assist:main",],},
)



#!/usr/bin/env python

import os, re, pathlib, json, copy, _thread
import docker
from sa import utils


def build(build_files):
    """Build and tag docker containers based on information provided in 'docker-info.json' files"""
    images_info=[]
    client = docker.from_env()

    for build_file in build_files:
        docker_file_data = utils.get_data_from_json(build_file)
        docker_build_info = utils.find_key_pairs(docker_file_data, ['img_name','file','tags','version','docker_registries'])
        
        for build_info in docker_build_info:
            temp_dict = {}

            temp_dict['path'] = str(pathlib.Path(build_info['file']).parent.resolve())
            temp_dict['dockerfile'] = str(pathlib.Path(build_info['file']).resolve())
            temp_dict['pull'] = True
            # TODO: add a way to process CLI options with variables, or from ENV variables
            temp_dict['buildargs'] = build_info['build_args']
            
            # add versions to tags
            for i in range(len(build_info['tags'])):
                build_info['tags'][i] = ''.join([build_info['tags'][i],'-',build_info['version']])

            # add default tag and registry if not preset
            build_info['docker_registries'].append('')
            build_info['tags'].append(build_info['version'])
            build_info['tags'].append('latest')

            # dedup lists
            build_info['docker_registries'] = list(set(build_info['docker_registries']))
            build_info['tags'] = list(set(build_info['tags']))

            # add image name to tag
            for i in range(len(build_info['tags'])):
                build_info['tags'][i] = ''.join([build_info['img_name'],':',build_info['tags'][i]])

            # implement tags and remotes
            for registry in build_info['docker_registries']:
                for tag in build_info['tags']:
                    temp_dict['tag'] = tag
                    if registry != '':
                        temp_dict['tag'] = ''.join([registry,'/',tag])
                    images_info.append(copy.deepcopy(temp_dict))

    for image_info in images_info:
        # Build and tag all images
        results = client.images.build(**image_info)
        for logs in results[-1]:
            for k,v in logs.items():
                if isinstance(v, str):
                    print(v.strip())
                    continue
                print(v)
                
        #_thread.start_new_thread(client.images.build,**image_info)


def push():
    pass



#!/usr/bin/env python

import sys
# add current directory to path
sys.path.insert(0, '')

from sa import _version, repo as rp, versioning, docker_api
import os, argparse, git

def main():
    # initialize argparser
    parser = argparse.ArgumentParser()
    # initialize sub argument parsers
    subparsers = parser.add_subparsers(dest='command')

    parser.add_argument('--version', action='version',
                    version='%(prog)s {version}'.format(version=_version.__version__))

    parser.add_argument('-gu','--gitusername',
                    dest='git_username',
                    type=str,
                    help='git username to use during commits, defaults to \'Jenkins\'')
    parser.add_argument('-ge','--gituseremail',
                    dest='git_useremail',
                    type=str,
                    help='git email to use during commits, defaults to \'jenkins@noreply.com\'')


    # git
    parser_git = subparsers.add_parser('git')
    subparser_git = parser_git.add_subparsers(dest='git_subcommand')
    # git push 
    git_subparser_push = subparser_git.add_parser('push')
    add_required_creds_args(git_subparser_push)
    add_required_repo_args(git_subparser_push)
    # git checkprevcommit
    git_subparser_checkprevcommit = subparsers.add_parser('checkprevcommit')
    add_required_repo_args(git_subparser_checkprevcommit)


    # version
    parser_version = subparsers.add_parser('version')
    subparser_version = parser_version.add_subparsers(dest='version_subcommand')
    # version get
    version_subparser_get = subparser_version.add_parser('get')
    add_optional_docker_arg(version_subparser_get)
    add_required_files_args(version_subparser_get)
    # version bump
    version_subparser_bump = subparser_version.add_parser('bump')
    add_required_repo_args(version_subparser_bump)
    add_optional_docker_arg(version_subparser_bump)
    add_required_files_args(version_subparser_bump)
    # TODO add flag to allow for custom provided versioning script


    # docker
    parser_docker = subparsers.add_parser('docker')
    subparser_docker = parser_docker.add_subparsers(dest='docker_subcommand')
    # docker build
    docker_subparser_build = subparser_docker.add_parser('build')
    add_required_files_args(docker_subparser_build)
    # docker geninfo
    #TODO: generate sample docker info json file


    # Process CLI arguments
    options = parser.parse_args()

    if options.command == 'git':

        if options.git_subcommand == 'push':
            rp.push(get_git_repo(options.repo_dir), options.username, options.password)
            exit()
        elif options.git_subcommand == 'checkprevcommit':
            #TODO: implement function
            rp.check_prev_commit(get_git_repo(options.repo_dir))

    elif options.command == 'version':

        if options.version_subcommand == 'bump':
            # Process all file paths provided and only include the valid files 
            rp.version_bump(get_git_repo(options.repo_dir), get_files(options.files), options)
            exit()
        elif options.version_subcommand == 'get':
            # Process all files paths provided and only include the valid files
            version_info = versioning.version_get(get_files(options.files), options)
            for file,version in version_info:
                print(''.join([file,' ', version]))
            exit()

    elif options.command == 'docker':

        if options.docker_subcommand == 'build':
            docker_api.build(get_files(options.files))
        elif options.docker_subcommand == 'push':
            docker_api.push()


def get_git_repo(repo_dir):
    # get/validate absolute path of provided repository directory
    abs_repo_path = ""
    if repo_dir != None:
        if os.path.isdir(repo_dir):
            abs_repo_path = os.path.abspath(repo_dir)
        else:
            # raise argparse.ArgumentTypeError(f"readable_dir:{repo_dir} is not a valid path")
            print(f'readable_dir:{abs_repo_path} is not a valid path')
            exit(1)
    
    # validate provided path is a git repository and setup for use 
    if (rp.is_git_repo(abs_repo_path)):
        repo = git.Repo(abs_repo_path)
        rp.reattach_head(repo)
    else:
        # raise Exception(f"{repo_dir} is not a valid git repository")
        print(f'{abs_repo_path} is not a valid git repository')
        exit(1)
    return repo


def get_files(infiles):
    files = []
    if infiles != None:
        for f in infiles:
            if os.path.isfile(f):
                files.append(os.path.abspath(f))
    
    if (len(files) == 0):
        print('no files provided, use \'-h|--help\' for more details'.format())
        exit(1)
    return files


def add_required_files_args(parser):
    parser.add_argument(
                    dest='files', 
                    type=str, 
                    nargs='*',
                    help='list of all files to be to be operated on')


def add_optional_docker_arg(parser):
    parser.add_argument('-d','--docker',
                    action='store_true',
                    help='indicate if provided files are docker-info.json files for versioning')


def add_required_creds_args(parser):
    parser.add_argument('-u','--username',
                    dest='username',
                    required=True,
                    help='username for logging into service')
    parser.add_argument('-p','--password',
                    dest='password',
                    required=True,
                    help='password for logging into service')


def add_required_repo_args(parser):
    parser.add_argument('-r','--repo',
                    dest='repo_dir',
                    type=str,
                    required=True,
                    help='specific path to git project files')


if __name__ == '__main__':
    sys.exit(main())